# pylint: disable=missing-class-docstring,no-self-use
from __future__ import annotations

import gc
import pickle
import unittest

import claripy


class TestPickle(unittest.TestCase):
    def test_pickle_ast(self):
        bz = claripy.backends.z3

        a = claripy.BVV(1, 32)
        b = claripy.BVS("x", 32, explicit_name=True)

        c = a + b
        assert bz.convert(c).__module__ == "z3.z3"
        assert str(bz.convert(c)), "1 + x"

        c_copy = pickle.loads(pickle.dumps(c, -1))
        assert c_copy is c
        assert bz.convert(c_copy).__module__ == "z3.z3"
        assert str(bz.convert(c_copy)) == "1 + x"

    def test_pickle_frontend(self):
        s = claripy.Solver()
        x = claripy.BVS("x", 32)

        s.add(x == 1)
        assert s.eval(x, 10), (1,)

        ss = pickle.dumps(s)
        del s

        gc.collect()

        s = pickle.loads(ss)
        assert s.eval(x, 10), (1,)

    def test_identity(self):
        a = claripy.BVV(1, 32)
        b = claripy.BVS("x", 32)
        c = a + b
        d = a + b * 50

        c_info = pickle.dumps(c)
        d_info = pickle.dumps(d)

        cc = pickle.loads(c_info)
        assert str(cc) == str(c)
        cd = pickle.loads(d_info)
        assert str(cd) == str(d)
        assert c.args[0] is d.args[0]

        s = claripy.Solver()
        x = claripy.BVS("x", 32)
        s.add(x == 3)
        s.finalize()
        ss = pickle.loads(pickle.dumps(s))
        assert str(s.constraints) == str(ss.constraints)
        assert str(s.variables) == str(ss.variables)

        s = claripy.SolverComposite()
        x = claripy.BVS("x", 32)
        s.add(x == 3)
        s.finalize()
        ss = pickle.loads(pickle.dumps(s))
        old_constraint_sets = [[hash(j) for j in k.constraints] for k in s._solver_list]
        new_constraint_sets = [[hash(j) for j in k.constraints] for k in ss._solver_list]
        assert old_constraint_sets == new_constraint_sets
        assert str(s.variables) == str(ss.variables)


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import functools
import itertools
import numbers

from claripy.bv import BVV

from .bool_result import BoolResult
from .errors import ClaripyVSAOperationError
from .strided_interval import StridedInterval
from .valueset import ValueSet

DEFAULT_MAX_CARDINALITY_WITHOUT_COLLAPSING = 256  # We don't collapse until there are more than this many SIs


def apply_on_each_si(f):
    @functools.wraps(f)
    def operator(self, o=None):
        if o is None:
            # This is an unary operator.
            new_si_set = set()

            for a in self._si_set:
                new_si_set.add(getattr(a, f.__name__)())

            ret = DiscreteStridedIntervalSet(bits=self.bits, si_set=new_si_set)
            return ret.normalize()

        if isinstance(o, DiscreteStridedIntervalSet):
            # We gotta apply the operation on each single object
            new_si_set = set()

            for a in self._si_set:
                for b in o._si_set:
                    new_si_set.add(getattr(a, f.__name__)(b))

            ret = DiscreteStridedIntervalSet(bits=self.bits, si_set=new_si_set)
            return ret.normalize()

        elif isinstance(o, StridedInterval | numbers.Number | BVV):
            new_si_set = set()
            for si in self._si_set:
                new_si_set.add(getattr(si, f.__name__)(o))

            ret = DiscreteStridedIntervalSet(bits=self.bits, si_set=new_si_set)
            return ret.normalize()

        else:
            raise ClaripyVSAOperationError(f"Unsupported operand type {type(o)}")

    return operator


def convert_operand_to_si(f):
    @functools.wraps(f)
    def converter(self, o):
        if isinstance(o, BVV):
            o = o.value
        if isinstance(o, numbers.Number):
            o = StridedInterval(bits=self.bits, stride=0, lower_bound=o, upper_bound=o)

        return f(self, o)

    return converter


def collapse_operand(f):
    @functools.wraps(f)
    def collapser(self, o):
        if isinstance(o, DiscreteStridedIntervalSet):
            return f(self, o.collapse())
        else:
            return f(self, o)

    return collapser


dsis_id_ctr = itertools.count()


class DiscreteStridedIntervalSet(StridedInterval):
    """
    A DiscreteStridedIntervalSet represents one or more discrete StridedInterval instances.
    """

    def __init__(self, name=None, bits=0, si_set=None, max_cardinality=None):
        if name is None:
            name = "DSIS_%d" % next(dsis_id_ctr)

        # Initialize the set for strided intervals
        if si_set is not None and len(si_set):
            self._si_set = si_set

        else:
            self._si_set = set()

        self._max_cardinality = (
            DEFAULT_MAX_CARDINALITY_WITHOUT_COLLAPSING if max_cardinality is None else max_cardinality
        )

        StridedInterval.__init__(self, name=name, bits=bits)

        # Update lower_bound and upper_bound
        for si in self._si_set:
            self._update_bounds(si)
            self._update_bits(si)

    #
    # Properties
    #

    def __repr__(self):
        representatives = ", ".join([i.__repr__() for i in list(self._si_set)[:5]])
        if self.number_of_values > 5:
            representatives += ", ..."

        return "%s<%d>(%d){%s}" % (self._name, self._bits, self.number_of_values, representatives)

    @property
    def cardinality(self):
        """
        This is an over-approximation of the cardinality of this DSIS.

        :return:
        """
        cardinality = 0
        for si in self._si_set:
            cardinality += si.cardinality
        return cardinality

    @property
    def number_of_values(self):
        return len(self._si_set)

    @property
    def stride(self):
        return self.collapse().stride

    #
    # Special methods
    #

    def should_collapse(self):
        return self.cardinality > self._max_cardinality

    def collapse(self):
        """
        Collapse into a StridedInterval instance.

        :return: A new StridedInterval instance.
        """

        if self.cardinality:
            r = None
            for si in self._si_set:
                r = r._union(si) if r is not None else si

            return r

        else:
            # This is an empty StridedInterval...
            return StridedInterval.empty(self._bits)

    def normalize(self):
        """
        Return the collapsed object if ``should_collapse()`` is True, otherwise return self.

        :return: A DiscreteStridedIntervalSet object.
        """
        if self.should_collapse():
            return self.collapse()
        elif self.number_of_values == 1:
            return next(iter(self._si_set))
        else:
            for si in self._si_set:
                self._update_bits(si)
            return self

    def copy(self):
        copied = DiscreteStridedIntervalSet(
            bits=self._bits, si_set=self._si_set.copy(), max_cardinality=self._max_cardinality
        )

        return copied

    def __hash__(self):
        return id(self)  # ...not sure how to do this. these objects are mutable.

    #
    # Operations
    #

    # Logical operations

    @convert_operand_to_si
    @collapse_operand
    def __eq__(self, o):
        """
        Operation ==

        :param o:   The other operand.
        :return:    An instance of BoolResult.
        """

        return self.collapse() == o

    @convert_operand_to_si
    @collapse_operand
    def __ne__(self, o):
        """
        Operation !=

        :param o:   The other operand.
        :return:    An instance of BoolResult.
        """

        return self.collapse() != o

    @convert_operand_to_si
    @collapse_operand
    def __gt__(self, o):
        """
        Operation >

        :param o:   The other operand.
        :return:    An instance of BoolResult.
        """

        return self.collapse() > o

    @convert_operand_to_si
    @collapse_operand
    def __le__(self, o):
        """
        Operation <=

        :param o:   The other operand.
        :return:    An instance of BoolResult.
        """

        return self.collapse() <= o

    @convert_operand_to_si
    @collapse_operand
    def __lt__(self, o):
        """
        Operation <

        :param o:   The other operand.
        :return:    An instance of BoolResult.
        """
        return self.collapse() < o

    # Bitwise operations

    @convert_operand_to_si
    @apply_on_each_si
    def __and__(self, o):
        """
        Operation &

        :param o:   The other operand.
        :return:    An instance of DiscreteStridedIntervalSet.
        """

    def __rand__(self, o):
        return self.__and__(o)

    @convert_operand_to_si
    @apply_on_each_si
    def __or__(self, o):
        """
        Operation |

        :param o:   The other operand.
        :return:    An instance of DiscreteStridedIntervalSet.
        """

    def __ror__(self, o):
        return self.__or__(o)

    @convert_operand_to_si
    @apply_on_each_si
    def __xor__(self, o):
        """
        Operation ^

        :param o:   The other operand.
        :return:    An instance of DiscreteStridedIntervalSet.
        """

    def __rxor__(self, o):
        return self.__xor__(o)

    def __neg__(self):
        """
        Operation ~

        :return: The negated value.
        """
        new_si_set = set()
        for si in self._si_set:
            new_si_set.add(~si)

        r = DiscreteStridedIntervalSet(bits=self._bits, si_set=new_si_set)
        return r.normalize()

    def __invert__(self):
        """
        Operation ~

        :return: The negated value.
        """
        return self.__neg__()

    @apply_on_each_si
    def __lshift__(self, o):
        """
        Operation <<

        :param o:   The other operand.
        :return:    An instance of DiscreteStridedIntervalSet.
        """

    @apply_on_each_si
    def __rshift__(self, o):
        """
        Operation >>

        :param o:   The other operand.
        :return:    An instance of DiscreteStridedIntervalSet.
        """

    @apply_on_each_si
    def concat(self, b):
        """
        Operation concat

        :param b:   The other operand to concatenate with.
        :return:    The concatenated value.
        """

    def extract(self, high_bit, low_bit):
        """
        Operation extract

        :param high_bit:    The highest bit to begin extraction.
        :param low_bit:     The lowest bit to end extraction.
        :return:            Extracted bits.
        """
        # TODO: This method can be optimized

        ret = set()
        bits = high_bit - low_bit + 1

        for si in self._si_set:
            ret.add(si.extract(high_bit, low_bit))

        if len(ret) > 1:
            return DiscreteStridedIntervalSet(bits=bits, si_set=ret)

        else:
            return next(iter(ret))

    # Arithmetic operations

    @convert_operand_to_si
    @apply_on_each_si
    def __add__(self, o):
        """
        Operation +

        :param o:   The other operand.
        :return:
        """

    def __radd__(self, o):
        return self.__add__(o)

    @convert_operand_to_si
    @apply_on_each_si
    def __sub__(self, o):
        """
        Operation -

        :param o:   The other operand.
        :return:
        """

    def __rsub__(self, o):
        return self.__sub__(o)

    @convert_operand_to_si
    @apply_on_each_si
    def __floordiv__(self, o):
        """
        Operation /

        :param o:   The other operand.
        :return:
        """

    def __truediv__(self, o):
        return self.__floordiv__(o)  # floats not welcome

    def __rfloordiv__(self, o):
        return self.__floordiv__(o)

    def __rtruediv__(self, o):
        return self.__rfloordiv__(o)

    @convert_operand_to_si
    @apply_on_each_si
    def __mod__(self, o):
        """
        Operation %

        :param o:   The other operand.
        :return:
        """

    def __rmod__(self, o):
        return self.__mod__(o)

    # Evaluation

    def eval(self, n, signed=False):
        """

        :param n:
        :param signed:
        :return:
        """

        # FIXME: "signed" is silently ignored now

        ret = set()

        for si in self._si_set:
            ret |= set(si.eval(n))
            if len(ret) >= n:
                break

        return list(ret)[:n]

    # Set operations

    def union(self, b):
        if isinstance(b, DiscreteStridedIntervalSet):
            return self._union_with_dsis(b)

        elif isinstance(b, StridedInterval):
            if not b.is_empty:
                return self._union_with_si(b)
            else:
                return self.copy()

        elif isinstance(b, ValueSet):
            return b.union(self)

        else:
            raise ClaripyVSAOperationError(f"Unsupported operand type {type(b)} for operation union.")

    def intersection(self, b):
        if isinstance(b, DiscreteStridedIntervalSet):
            return self._intersection_with_dsis(b)

        elif isinstance(b, StridedInterval):
            return self._intersection_with_si(b)

        else:
            raise ClaripyVSAOperationError(f"Unsupported operand type {type(b)} for operation intersection.")

    # Other operations

    @apply_on_each_si
    def reverse(self):
        """
        Operation Reverse

        :return: None
        """

    @apply_on_each_si
    def sign_extend(self, new_length):
        """
        Operation SignExt

        :param new_length:  The length to extend to.
        :return:            SignExtended value.
        """

    @apply_on_each_si
    def zero_extend(self, new_length):
        """
        Operation ZeroExt

        :param new_length:  The length to extend to.
        :return:            ZeroExtended value.
        """

    @collapse_operand
    def widen(self, b):
        """
        Widening operator.

        :param b:   The other operand.
        :return:    The widened result.
        """
        return self.collapse().widen(b)

    #
    # Private methods
    #

    def _union_with_si(self, si):
        """
        Union with another StridedInterval.

        :param si:
        :return:
        """

        dsis = self.copy()
        for si_ in dsis._si_set:
            if BoolResult.is_true(si_ == si):
                return dsis

        dsis._si_set.add(si)
        dsis._update_bounds(si)

        return dsis.normalize()

    def _union_with_dsis(self, dsis):
        """
        Union with another DiscreteStridedIntervalSet.

        :param dsis:
        :return:
        """

        copied = self.copy()

        for a in dsis._si_set:
            copied = copied.union(a)

        if isinstance(copied, DiscreteStridedIntervalSet):
            copied._update_bounds(dsis)

        return copied.normalize()

    def _intersection_with_si(self, si):
        """
        Intersection with another :class:`StridedInterval`.

        :param si: The other operand
        :return:
        """

        new_si_set = set()
        for si_ in self._si_set:
            r = si_.intersection(si)
            new_si_set.add(r)

        if len(new_si_set):
            ret = DiscreteStridedIntervalSet(bits=self.bits, si_set=new_si_set)

            if ret.should_collapse():
                return ret.collapse()
            else:
                return ret

        else:
            # There is no intersection between two operands
            return StridedInterval.empty(self.bits)

    def _intersection_with_dsis(self, dsis):
        """
        Intersection with another :class:`DiscreteStridedIntervalSet`.

        :param dsis:    The other operand.
        :return:
        """

        new_si_set = set()
        for si in dsis._si_set:
            r = self._intersection_with_si(si)

            if isinstance(r, StridedInterval):
                if not r.is_empty:
                    new_si_set.add(r)

            else:  # r is a DiscreteStridedIntervalSet
                new_si_set |= r._si_set

        if len(new_si_set):
            ret = DiscreteStridedIntervalSet(bits=self.bits, si_set=new_si_set)

            return ret.normalize()

        else:
            return StridedInterval.empty(self.bits)

    def _update_bounds(self, val):
        if not isinstance(val, StridedInterval):
            raise ClaripyVSAOperationError(f"Unsupported operand type {type(val)}.")

        if val._lower_bound is not None and (self._lower_bound is None or val.lower_bound < self._lower_bound):
            self._lower_bound = val.lower_bound

        if val._upper_bound is not None and (self._upper_bound is None or val.upper_bound > self._upper_bound):
            self._upper_bound = val.upper_bound

    def _update_bits(self, val):
        if not isinstance(val, StridedInterval):
            raise ClaripyVSAOperationError(f"Unsupported operand type {type(val)}.")

        self._bits = val.bits



from __future__ import annotations

from . import backends, frontend_mixins, frontends


class Solver(
    frontend_mixins.ConcreteHandlerMixin,
    frontend_mixins.EagerResolutionMixin,
    frontend_mixins.ConstraintFilterMixin,
    frontend_mixins.ConstraintDeduplicatorMixin,
    frontend_mixins.SimplifySkipperMixin,
    frontend_mixins.SatCacheMixin,
    frontend_mixins.ModelCacheMixin,
    frontend_mixins.ConstraintExpansionMixin,
    frontend_mixins.SimplifyHelperMixin,
    frontends.FullFrontend,
):
    def __init__(self, backend=backends.z3, **kwargs):
        super().__init__(backend, **kwargs)


class SolverCacheless(
    frontend_mixins.ConcreteHandlerMixin,
    frontend_mixins.EagerResolutionMixin,
    frontend_mixins.ConstraintFilterMixin,
    frontend_mixins.ConstraintDeduplicatorMixin,
    frontend_mixins.SimplifySkipperMixin,
    frontends.FullFrontend,
):
    def __init__(self, backend=backends.z3, **kwargs):
        super().__init__(backend, **kwargs)


class SolverReplacement(
    frontend_mixins.ConcreteHandlerMixin,
    frontend_mixins.ConstraintDeduplicatorMixin,
    frontends.ReplacementFrontend,
):
    def __init__(self, actual_frontend=None, **kwargs):
        actual_frontend = Solver() if actual_frontend is None else actual_frontend
        super().__init__(actual_frontend, **kwargs)


class SolverHybrid(
    frontend_mixins.ConcreteHandlerMixin,
    frontend_mixins.EagerResolutionMixin,
    frontend_mixins.ConstraintFilterMixin,
    frontend_mixins.ConstraintDeduplicatorMixin,
    frontend_mixins.SimplifySkipperMixin,
    # TODO: frontend_mixins.ConstraintExpansionMixin,
    frontends.HybridFrontend,
):
    def __init__(
        self,
        exact_frontend=None,
        approximate_frontend=None,
        complex_auto_replace=True,
        replace_constraints=True,
        track=False,
        approximate_first=False,
        **kwargs,
    ):
        exact_frontend = Solver(track=track) if exact_frontend is None else exact_frontend
        approximate_frontend = (
            SolverReplacement(
                actual_frontend=SolverVSA(),
                complex_auto_replace=complex_auto_replace,
                replace_constraints=replace_constraints,
            )
            if approximate_frontend is None
            else approximate_frontend
        )
        super().__init__(exact_frontend, approximate_frontend, approximate_first=approximate_first, **kwargs)


class SolverVSA(
    frontend_mixins.ConcreteHandlerMixin,
    frontend_mixins.ConstraintFilterMixin,
    frontends.LightFrontend,
):
    def __init__(self, **kwargs):
        super().__init__(backends.vsa, **kwargs)


class SolverConcrete(
    frontend_mixins.ConcreteHandlerMixin,
    frontend_mixins.ConstraintFilterMixin,
    frontends.LightFrontend,
):
    def __init__(self, **kwargs):
        super().__init__(backends.concrete, **kwargs)


class SolverStrings(
    # TODO: Figure ot if we need to use all these mixins
    frontend_mixins.ConcreteHandlerMixin,
    frontend_mixins.ConstraintFilterMixin,
    frontend_mixins.ConstraintDeduplicatorMixin,
    frontend_mixins.EagerResolutionMixin,
    frontend_mixins.EvalStringsToASTsMixin,
    frontends.FullFrontend,
):
    def __init__(self, *args, backend=backends.z3, **kwargs):
        super().__init__(backend, *args, **kwargs)


#
# Composite solving
#


class SolverCompositeChild(
    frontend_mixins.ConstraintDeduplicatorMixin,
    frontend_mixins.SatCacheMixin,
    frontend_mixins.SimplifySkipperMixin,
    frontend_mixins.ModelCacheMixin,
    frontends.FullFrontend,
):
    def __init__(self, backend=backends.z3, **kwargs):
        super().__init__(backend, **kwargs)

    def __repr__(self):
        return "<SolverCompositeChild with %d variables>" % len(self.variables)


class SolverComposite(
    frontend_mixins.ConcreteHandlerMixin,
    frontend_mixins.EagerResolutionMixin,
    frontend_mixins.ConstraintFilterMixin,
    frontend_mixins.ConstraintDeduplicatorMixin,
    frontend_mixins.SatCacheMixin,
    frontend_mixins.SimplifySkipperMixin,
    frontend_mixins.SimplifyHelperMixin,
    frontend_mixins.ConstraintExpansionMixin,
    frontend_mixins.CompositedCacheMixin,
    frontends.CompositeFrontend,
):
    def __init__(self, template_solver=None, track=False, template_solver_string=None, **kwargs):
        template_solver = SolverCompositeChild(track=track) if template_solver is None else template_solver
        template_solver_string = (
            SolverCompositeChild(track=track, backend=backends.z3)
            if template_solver_string is None
            else template_solver_string
        )
        super().__init__(template_solver, template_solver_string, track=track, **kwargs)

    def __repr__(self):
        return "<SolverComposite %x, %d children>" % (id(self), len(self._solver_list))



from typing import Any, overload

from claripy.bv import BVV as ConcreteBVV
from claripy.fp import RM, FSort

from .bits import Bits
from .bool import Bool
from .fp import FP

CoerceBV = int | "BV"
Incomplete = Any

class BV(Bits):
    def chop(self, bits: int = 1) -> list[BV]: ...
    def __getitem__(self, rng: int | slice) -> BV: ...
    def get_byte(self, index: int) -> BV: ...
    def get_bytes(self, index: int, size: int) -> BV: ...
    def zero_extend(self, n: int) -> BV: ...
    def sign_extend(self, n: int) -> BV: ...
    def concat(self, *args: BV) -> BV: ...
    @staticmethod
    def _from_int(like, value: int) -> BV: ...
    @staticmethod
    def _from_Bool(like, value: Bool) -> BV: ...
    @staticmethod
    def _from_bytes(like, value: bytes) -> BV: ...
    @staticmethod
    def _from_str(like, value: str) -> BV: ...
    @staticmethod
    def _from_BVV(like, value: ConcreteBVV) -> BV: ...
    def val_to_fp(self, sort: FSort, signed: bool = True, rm: RM | None = None) -> FP: ...
    def raw_to_fp(self) -> FP: ...
    def raw_to_bv(self) -> BV: ...
    def to_bv(self) -> BV: ...
    def __add__(self, other: CoerceBV) -> BV: ...
    def __radd__(self, other: CoerceBV) -> BV: ...
    def __floordiv__(self, other: CoerceBV) -> BV: ...
    def __rfloordiv__(self, other: CoerceBV) -> BV: ...
    def __truediv__(self, other: CoerceBV) -> BV: ...
    def __rtruediv__(self, other: CoerceBV) -> BV: ...
    def __mul__(self, other: CoerceBV) -> BV: ...
    def __rmul__(self, other: CoerceBV) -> BV: ...
    def __sub__(self, other: CoerceBV) -> BV: ...
    def __rsub__(self, other: CoerceBV) -> BV: ...
    def __pow__(self, other: CoerceBV) -> BV: ...
    def __rpow__(self, other: CoerceBV) -> BV: ...
    def __mod__(self, other: CoerceBV) -> BV: ...
    def __rmod__(self, other: CoerceBV) -> BV: ...
    def SDiv(self, other: CoerceBV) -> BV: ...
    def SMod(self, other: CoerceBV) -> BV: ...
    def __or__(self, other: CoerceBV) -> BV: ...
    def __ror__(self, other: CoerceBV) -> BV: ...
    def __and__(self, other: CoerceBV) -> BV: ...
    def __rand__(self, other: CoerceBV) -> BV: ...
    def __xor__(self, other: CoerceBV) -> BV: ...
    def __rxor__(self, other: CoerceBV) -> BV: ...
    def __lshift__(self, other: CoerceBV) -> BV: ...
    def __rlshift__(self, other: CoerceBV) -> BV: ...
    def __rshift__(self, other: CoerceBV) -> BV: ...
    def __rrshift__(self, other: CoerceBV) -> BV: ...
    def LShR(self, other: CoerceBV) -> BV: ...
    def __neg__(self) -> BV: ...
    def __pos__(self) -> BV: ...
    def __abs__(self) -> BV: ...
    def __invert__(self) -> BV: ...
    def __eq__(self, other: CoerceBV) -> Bool: ...
    def __ne__(self, other: CoerceBV) -> Bool: ...
    def __ge__(self, other: CoerceBV) -> Bool: ...
    def __le__(self, other: CoerceBV) -> Bool: ...
    def __gt__(self, other: CoerceBV) -> Bool: ...
    def __lt__(self, other: CoerceBV) -> Bool: ...
    def SLT(self, other: CoerceBV) -> Bool: ...
    def SLE(self, other: CoerceBV) -> Bool: ...
    def SGT(self, other: CoerceBV) -> Bool: ...
    def SGE(self, other: CoerceBV) -> Bool: ...
    def ULT(self, other: CoerceBV) -> Bool: ...
    def ULE(self, other: CoerceBV) -> Bool: ...
    def UGT(self, other: CoerceBV) -> Bool: ...
    def UGE(self, other: CoerceBV) -> Bool: ...
    def Extract(self, start: int, end: int) -> BV: ...
    def Concat(self, other: BV) -> BV: ...
    @property
    def reversed(self) -> BV: ...

    # weh
    def union(self, other: BV) -> BV: ...
    def widen(self, other: BV) -> BV: ...
    def intersection(self, other: BV) -> BV: ...
    @property
    def concrete_value(self) -> int: ...

def BVS(
    name: str,
    size: int,
    min=None,
    max=None,
    stride=None,
    uninitialized=False,  # pylint:disable=redefined-builtin
    explicit_name=None,
    discrete_set=False,
    discrete_set_max_card=None,
    **kwargs,
) -> BV: ...
@overload
def BVV(value: int, size: int = ..., **kwargs) -> BV: ...
@overload
def BVV(value: bytes, **kwargs) -> BV: ...
def SI(
    name: str | None = ...,
    bits: int = ...,
    lower_bound: int | None = ...,
    upper_bound: int | None = ...,
    stride: int | None = ...,
    to_conv: Incomplete | None = ...,
    explicit_name: bool | None = ...,
    discrete_set: bool = ...,
    discrete_set_max_card: int | None = ...,
): ...
def TSI(bits, name: str | None = ..., uninitialized: bool = ..., explicit_name: bool | None = ...): ...
def ESI(bits, **kwargs): ...
def ValueSet(
    bits,
    region: str | None = ...,
    region_base_addr: int | None = ...,
    value: int | None = ...,
    name: str | None = ...,
    val: Incomplete | None = ...,
): ...

VS = ValueSet

def DSIS(
    name: Incomplete | None = ...,
    bits: int = ...,
    lower_bound: Incomplete | None = ...,
    upper_bound: Incomplete | None = ...,
    stride: Incomplete | None = ...,
    explicit_name: Incomplete | None = ...,
    to_conv: Incomplete | None = ...,
    max_card: Incomplete | None = ...,
): ...
def ULT(a: BV | int, b: BV | int) -> Bool: ...
def ULE(a: BV | int, b: BV | int) -> Bool: ...
def UGT(a: BV | int, b: BV | int) -> Bool: ...
def UGE(a: BV | int, b: BV | int) -> Bool: ...
def SLT(a: BV | int, b: BV | int) -> Bool: ...
def SLE(a: BV | int, b: BV | int) -> Bool: ...
def SGT(a: BV | int, b: BV | int) -> Bool: ...
def SGE(a: BV | int, b: BV | int) -> Bool: ...
def SDiv(a: BV | int, b: BV | int) -> BV: ...
def SMod(a: BV | int, b: BV | int) -> BV: ...
def LShR(a: BV | int, b: BV | int) -> BV: ...
def SignExt(a: int, b: BV | int) -> BV: ...
def ZeroExt(a: int, b: BV | int) -> BV: ...
def Extract(a: int, b: int, BV) -> BV: ...
def Concat(*a: BV) -> BV: ...
def RotateLeft(a: BV | int, b: BV | int) -> BV: ...
def RotateRight(a: BV | int, b: BV | int) -> BV: ...
def Reverse(a: BV | int) -> BV: ...
def union(a: BV | int, b: BV | int) -> BV: ...
def widen(a: BV | int, b: BV | int) -> BV: ...
def intersection(a: BV | int, b: BV | int) -> BV: ...



from __future__ import annotations


class ConstraintDeduplicatorMixin:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._constraint_hashes = set()

    def _blank_copy(self, c):
        super()._blank_copy(c)
        c._constraint_hashes = set()

    def _copy(self, c):
        super()._copy(c)
        c._constraint_hashes = set(self._constraint_hashes)

    def __getstate__(self):
        return self._constraint_hashes, super().__getstate__()

    def __setstate__(self, s):
        self._constraint_hashes, base_state = s
        super().__setstate__(base_state)

    def simplify(self):
        added = super().simplify()
        # we only add to the constraint hashes because we want to
        # prevent previous (now simplified) constraints from
        # being re-added
        self._constraint_hashes.update(map(hash, added))
        return added

    def _add(self, constraints, invalidate_cache=True):
        filtered = tuple(c for c in constraints if hash(c) not in self._constraint_hashes)
        if len(filtered) == 0:
            return filtered

        added = super()._add(filtered, invalidate_cache=invalidate_cache)
        self._constraint_hashes.update(map(hash, added))
        return added



from __future__ import annotations

from claripy import String, StringV


class EvalStringsToASTsMixin:
    def eval_to_ast(self, e, n, extra_constraints=(), exact=None):
        if type(e) is String:
            return [
                StringV(
                    v,
                )
                for v in self.eval(e, n, extra_constraints=extra_constraints, exact=exact)
            ]

        return super().eval_to_ast(e, n, extra_constraints=extra_constraints, exact=None)



from __future__ import annotations

import functools
import logging
import numbers
import operator
from functools import reduce

from claripy.ast.base import Base
from claripy.backends.backend import Backend
from claripy.balancer import Balancer
from claripy.errors import BackendError
from claripy.operations import backend_operations_vsa_compliant, expression_set_operations
from claripy.vsa import (
    BoolResult,
    CreateStridedInterval,
    DiscreteStridedIntervalSet,
    FalseResult,
    RegionAnnotation,
    StridedInterval,
    TrueResult,
    ValueSet,
)

l = logging.getLogger("claripy.backends.backend_vsa")


def arg_filter(f):
    @functools.wraps(f)
    def filter(*args):  # pylint:disable=redefined-builtin
        if isinstance(args[0], numbers.Number):  # pylint:disable=unidiomatic-typecheck
            raise BackendError(f"Unsupported argument type {type(args[0])}")
        return f(*args)

    return filter


def normalize_arg_order(f):
    @functools.wraps(f)
    def normalizer(*args):
        if len(args) != 2:
            raise BackendError("Unsupported arguments number %d" % len(args))

        if type(args[0]) not in {
            StridedInterval,
            DiscreteStridedIntervalSet,
            ValueSet,
        }:  # pylint:disable=unidiomatic-typecheck
            if type(args[1]) not in {
                StridedInterval,
                DiscreteStridedIntervalSet,
                ValueSet,
            }:  # pylint:disable=unidiomatic-typecheck
                raise BackendError("Unsupported arguments")
            args = [args[1], args[0]]

        return f(*args)

    return normalizer


def convert_args(f):
    @functools.wraps(f)
    def converter(self, ast):
        raw_args = []
        for i in range(len(ast.args)):
            # It's not reversed
            raw_args.append(ast.args[i])

        for i in range(len(raw_args)):
            raw_args[i] = self.convert(raw_args[i])

        normalized = ast.swap_args(raw_args)
        ret = f(self, normalized)

        return ret

    return converter


class BackendVSA(Backend):
    def __init__(self):
        Backend.__init__(self)
        # self._make_raw_ops(set(expression_operations) - set(expression_set_operations), op_module=BackendVSA)
        self._make_expr_ops(set(expression_set_operations), op_class=self)
        self._make_raw_ops(set(backend_operations_vsa_compliant), op_module=BackendVSA)

        self._op_raw["Reverse"] = BackendVSA.Reverse
        self._op_raw["If"] = self.If
        self._op_expr["BVV"] = self.BVV
        self._op_expr["BoolV"] = self.BoolV
        self._op_expr["BVS"] = self.BVS

        # reduceable
        self._op_raw["__add__"] = self._op_add
        self._op_raw["__sub__"] = self._op_sub
        self._op_raw["__mul__"] = self._op_mul
        self._op_raw["__or__"] = self._op_or
        self._op_raw["__xor__"] = self._op_xor
        self._op_raw["__and__"] = self._op_and
        self._op_raw["__mod__"] = self._op_mod

    @staticmethod
    def _op_add(*args):
        return reduce(operator.__add__, args)

    @staticmethod
    def _op_sub(*args):
        return reduce(operator.__sub__, args)

    @staticmethod
    def _op_mul(*args):
        return reduce(operator.__mul__, args)

    @staticmethod
    def _op_or(*args):
        return reduce(operator.__or__, args)

    @staticmethod
    def _op_xor(*args):
        return reduce(operator.__xor__, args)

    @staticmethod
    def _op_and(*args):
        return reduce(operator.__and__, args)

    @staticmethod
    def _op_mod(*args):
        return reduce(operator.__mod__, args)

    def convert(self, expr):
        return Backend.convert(self, expr.ite_excavated if isinstance(expr, Base) else expr)

    def _convert(self, a):
        if isinstance(a, numbers.Number):
            return a
        elif isinstance(a, bool):
            return TrueResult() if a else FalseResult()
        if isinstance(a, StridedInterval | DiscreteStridedIntervalSet | ValueSet):
            return a
        if isinstance(a, BoolResult):
            return a

        # Not supported
        raise BackendError

    def _eval(self, expr, n, extra_constraints=(), solver=None, model_callback=None):
        if isinstance(expr, StridedInterval | ValueSet):
            return expr.eval(n)
        elif isinstance(expr, BoolResult):
            return expr.value
        else:
            raise BackendError(f"Unsupported type {type(expr)}")

    def _min(self, expr, extra_constraints=(), signed=False, solver=None, model_callback=None):
        # TODO: signed min
        if isinstance(expr, StridedInterval):
            if expr.is_top:
                # TODO: Return
                return 0

            return expr.min

        elif isinstance(expr, ValueSet):
            return expr.min

        else:
            raise BackendError(f"Unsupported expr type {type(expr)}")

    def _max(self, expr, extra_constraints=(), signed=False, solver=None, model_callback=None):
        # TODO: signed max
        if isinstance(expr, StridedInterval):
            if expr.is_top:
                # TODO:
                return StridedInterval.max_int(expr.bits)

            return expr.max

        elif isinstance(expr, ValueSet):
            return expr.max

        else:
            raise BackendError(f"Unsupported expr type {type(expr)}")

    def _solution(self, obj, v, extra_constraints=(), solver=None, model_callback=None):
        if isinstance(obj, BoolResult):
            return len(set(v.value) & set(obj.value)) > 0

        if isinstance(obj, StridedInterval):
            return not obj.intersection(v).is_empty

        if isinstance(obj, ValueSet):
            return any(not si.intersection(v).is_empty for _, si in obj.items())

        raise NotImplementedError(type(obj).__name__)

    def _has_true(self, o, extra_constraints=(), solver=None, model_callback=None):
        return BoolResult.has_true(o)

    def _has_false(self, o, extra_constraints=(), solver=None, model_callback=None):
        return BoolResult.has_false(o)

    def _is_true(self, o, extra_constraints=(), solver=None, model_callback=None):
        return BoolResult.is_true(o)

    def _is_false(self, o, extra_constraints=(), solver=None, model_callback=None):
        return BoolResult.is_false(o)

    #
    # Backend Operations
    #

    def simplify(self, e):
        raise BackendError("nope")

    def _identical(self, a, b):
        if type(a) != type(b):  # noqa: E721
            return False
        return a.identical(b)

    def _unique(self, obj):  # pylint:disable=unused-argument,no-self-use
        if isinstance(obj, StridedInterval | ValueSet):
            return obj.unique
        else:
            raise BackendError(f"Not supported type of operand {type(obj)}")

    def _cardinality(self, a):  # pylint:disable=unused-argument,no-self-use
        return a.cardinality

    def name(self, a):
        if isinstance(a, StridedInterval):
            return a.name

        else:
            return None

    def apply_annotation(self, bo, annotation):
        """
        Apply an annotation on the backend object.

        :param BackendObject bo: The backend object.
        :param Annotation annotation: The annotation to be applied
        :return: A new BackendObject
        :rtype: BackendObject
        """

        # Currently we only support RegionAnnotation

        if not isinstance(annotation, RegionAnnotation):
            return bo

        if not isinstance(bo, ValueSet):
            # Convert it to a ValueSet first
            # Note that the original value is not kept at all. If you want to convert a StridedInterval to a ValueSet,
            # you gotta do the conversion by calling AST.annotate() from outside.
            bo = ValueSet.empty(bo.bits)

        return bo.apply_annotation(annotation)

    def BVV(self, ast):  # pylint:disable=unused-argument,no-self-use
        if ast.args[0] is None:
            return StridedInterval.empty(ast.args[1])
        else:
            return CreateStridedInterval(bits=ast.args[1], stride=0, lower_bound=ast.args[0], upper_bound=ast.args[0])

    @staticmethod
    def BoolV(ast):  # pylint:disable=unused-argument
        return TrueResult() if ast.args[0] else FalseResult()

    @staticmethod
    def And(a, *args):
        return reduce(operator.__and__, args, a)

    @staticmethod
    def Not(a):
        return ~a

    @staticmethod
    @normalize_arg_order
    def ULT(a, b):
        return a.ULT(b)

    @staticmethod
    @normalize_arg_order
    def ULE(a, b):
        return a.ULE(b)

    @staticmethod
    @normalize_arg_order
    def UGT(a, b):
        return a.UGT(b)

    @staticmethod
    @normalize_arg_order
    def UGE(a, b):
        return a.UGE(b)

    @staticmethod
    @normalize_arg_order
    def SLT(a, b):
        return a.SLT(b)

    @staticmethod
    @normalize_arg_order
    def SLE(a, b):
        return a.SLE(b)

    @staticmethod
    @normalize_arg_order
    def SGT(a, b):
        return a.SGT(b)

    @staticmethod
    @normalize_arg_order
    def SGE(a, b):
        return a.SGE(b)

    @staticmethod
    def BVS(ast):  # pylint:disable=unused-argument
        size = ast.size()
        name, mn, mx, stride, uninitialized, discrete_set, max_card = ast.args
        return CreateStridedInterval(
            name=name,
            bits=size,
            lower_bound=mn,
            upper_bound=mx,
            stride=stride,
            uninitialized=uninitialized,
            discrete_set=discrete_set,
            discrete_set_max_cardinality=max_card,
        )

    def If(self, cond, t, f):
        if not self.has_true(cond):
            return f
        elif not self.has_false(cond):
            return t
        else:
            return t.union(f)

    # TODO: Implement other operations!

    @staticmethod
    def Or(*args):
        first = args[0]
        others = args[1:]

        for o in others:
            first = first.union(o)
        return first

    @staticmethod
    def __rshift__(expr, shift_amount):  # pylint:disable=unexpected-special-method-signature
        return expr.__rshift__(shift_amount)

    @staticmethod
    def LShR(expr, shift_amount):
        return expr.LShR(shift_amount)

    @staticmethod
    def Concat(*args):
        ret = None
        for expr in args:
            if type(expr) not in {
                StridedInterval,
                DiscreteStridedIntervalSet,
                ValueSet,
            }:  # pylint:disable=unidiomatic-typecheck
                raise BackendError(f"Unsupported expr type {type(expr)}")

            ret = ret.concat(expr) if ret is not None else expr

        return ret

    @staticmethod
    def Extract(*args):
        low_bit = args[1]
        high_bit = args[0]
        expr = args[2]

        if type(expr) not in {
            StridedInterval,
            DiscreteStridedIntervalSet,
            ValueSet,
        }:  # pylint:disable=unidiomatic-typecheck
            raise BackendError(f"Unsupported expr type {type(expr)}")

        ret = expr.extract(high_bit, low_bit)

        return ret

    @staticmethod
    def SignExt(*args):
        new_bits = args[0]
        expr = args[1]

        if type(expr) not in {StridedInterval, DiscreteStridedIntervalSet}:  # pylint:disable=unidiomatic-typecheck
            raise BackendError(f"Unsupported expr type {type(expr)}")

        return expr.sign_extend(new_bits + expr.bits)

    @staticmethod
    def ZeroExt(*args):
        new_bits = args[0]
        expr = args[1]

        if type(expr) not in {StridedInterval, DiscreteStridedIntervalSet}:  # pylint:disable=unidiomatic-typecheck
            raise BackendError(f"Unsupported expr type {type(expr)}")

        return expr.zero_extend(new_bits + expr.bits)

    @staticmethod
    def Reverse(arg):
        if type(arg) not in {
            StridedInterval,
            DiscreteStridedIntervalSet,
            ValueSet,
        }:  # pylint:disable=unidiomatic-typecheck
            raise BackendError(f"Unsupported expr type {type(arg)}")

        return arg.reverse()

    @convert_args
    def union(self, ast):  # pylint:disable=unused-argument,no-self-use
        if len(ast.args) != 2:
            raise BackendError("Incorrect number of arguments (%d) passed to BackendVSA.union()." % len(ast.args))

        ret = ast.args[0].union(ast.args[1])

        if ret is NotImplemented:
            ret = ast.args[1].union(ast.args[0])

        return ret

    @convert_args
    def intersection(self, ast):  # pylint:disable=unused-argument,no-self-use
        if len(ast.args) != 2:
            raise BackendError(
                "Incorrect number of arguments (%d) passed to BackendVSA.intersection()." % len(ast.args)
            )

        ret = None

        for arg in ast.args:
            ret = arg if ret is None else ret.intersection(arg)
        return ret

    @convert_args
    def widen(self, ast):  # pylint:disable=unused-argument,no-self-use
        if len(ast.args) != 2:
            raise BackendError("Incorrect number of arguments (%d) passed to BackendVSA.widen()." % len(ast.args))

        ret = ast.args[0].widen(ast.args[1])
        if ret is NotImplemented:
            ret = ast.args[1].widen(ast.args[0])

        return ret

    @staticmethod
    def CreateTopStridedInterval(bits, name=None, uninitialized=False):  # pylint:disable=unused-argument,no-self-use
        return StridedInterval.top(bits, name, uninitialized=uninitialized)

    def constraint_to_si(self, expr):
        return Balancer(self, expr).compat_ret


BackendVSA.CreateStridedInterval = staticmethod(CreateStridedInterval)



# mixins for frontends
from .constraint_expansion_mixin import ConstraintExpansionMixin
from .model_cache_mixin import ModelCacheMixin
from .constraint_filter_mixin import ConstraintFilterMixin
from .eager_resolution_mixin import EagerResolutionMixin
from .constraint_deduplicator_mixin import ConstraintDeduplicatorMixin
from .concrete_handler_mixin import ConcreteHandlerMixin
from .debug_mixin import DebugMixin
from .solve_block_mixin import SolveBlockMixin
from .simplify_helper_mixin import SimplifyHelperMixin
from .simplify_skipper_mixin import SimplifySkipperMixin
from .composited_cache_mixin import CompositedCacheMixin
from .sat_cache_mixin import SatCacheMixin
from .eval_string_to_ast_mixin import EvalStringsToASTsMixin



# pylint: disable=missing-class-docstring,no-self-use
from __future__ import annotations

import unittest

import claripy


class TestBalancer(unittest.TestCase):
    def test_complex(self):
        guy_wide = claripy.widen(
            claripy.union(
                claripy.union(
                    claripy.union(claripy.BVV(0, 32), claripy.BVV(1, 32)),
                    claripy.union(claripy.BVV(0, 32), claripy.BVV(1, 32)) + claripy.BVV(1, 32),
                ),
                claripy.union(
                    claripy.union(claripy.BVV(0, 32), claripy.BVV(1, 32)),
                    claripy.union(claripy.BVV(0, 32), claripy.BVV(1, 32)) + claripy.BVV(1, 32),
                )
                + claripy.BVV(1, 32),
            ),
            claripy.union(
                claripy.union(
                    claripy.union(
                        claripy.union(claripy.BVV(0, 32), claripy.BVV(1, 32)),
                        claripy.union(claripy.BVV(0, 32), claripy.BVV(1, 32)) + claripy.BVV(1, 32),
                    ),
                    claripy.union(
                        claripy.union(claripy.BVV(0, 32), claripy.BVV(1, 32)),
                        claripy.union(claripy.BVV(0, 32), claripy.BVV(1, 32)) + claripy.BVV(1, 32),
                    )
                    + claripy.BVV(1, 32),
                ),
                claripy.union(
                    claripy.union(
                        claripy.union(claripy.BVV(0, 32), claripy.BVV(1, 32)),
                        claripy.union(claripy.BVV(0, 32), claripy.BVV(1, 32)) + claripy.BVV(1, 32),
                    ),
                    claripy.union(
                        claripy.union(claripy.BVV(0, 32), claripy.BVV(1, 32)),
                        claripy.union(claripy.BVV(0, 32), claripy.BVV(1, 32)) + claripy.BVV(1, 32),
                    )
                    + claripy.BVV(1, 32),
                )
                + claripy.BVV(1, 32),
            ),
        )
        guy_inc = guy_wide + claripy.BVV(1, 32)
        guy_zx = claripy.ZeroExt(32, guy_inc)

        s, r = claripy.balancer.Balancer(claripy.backends.vsa, guy_inc <= claripy.BVV(39, 32)).compat_ret
        assert s
        assert r[0][0] is guy_wide
        assert claripy.backends.vsa.min(r[0][1]) == 0
        assert set(claripy.backends.vsa.eval(r[0][1], 1000)) == {4294967295, *list(range(39))}

        s, r = claripy.balancer.Balancer(claripy.backends.vsa, guy_zx <= claripy.BVV(39, 64)).compat_ret
        assert r[0][0] is guy_wide
        assert claripy.backends.vsa.min(r[0][1]) == 0
        assert set(claripy.backends.vsa.eval(r[0][1], 1000)) == {4294967295, *list(range(39))}

    def test_simple(self):
        x = claripy.BVS("x", 32)
        s, r = claripy.balancer.Balancer(claripy.backends.vsa, x <= claripy.BVV(39, 32)).compat_ret
        assert s
        assert r[0][0] is x
        assert claripy.backends.vsa.min(r[0][1]) == 0
        assert claripy.backends.vsa.max(r[0][1]) == 39

        s, r = claripy.balancer.Balancer(claripy.backends.vsa, x + 1 <= claripy.BVV(39, 32)).compat_ret
        assert s
        assert r[0][0] is x
        all_vals = r[0][1]._model_vsa.eval(1000)
        assert len(all_vals)
        assert min(all_vals) == 0
        assert max(all_vals) == 4294967295
        all_vals.remove(4294967295)
        assert max(all_vals) == 38

    def test_widened(self):
        w = claripy.widen(claripy.BVV(1, 32), claripy.BVV(0, 32))
        s, r = claripy.balancer.Balancer(claripy.backends.vsa, w <= claripy.BVV(39, 32)).compat_ret
        assert s
        assert r[0][0] is w
        assert claripy.backends.vsa.min(r[0][1]) == 0
        assert claripy.backends.vsa.max(r[0][1]) == 1  # used to be 39, but that was a bug in the VSA widening

        s, r = claripy.balancer.Balancer(claripy.backends.vsa, w + 1 <= claripy.BVV(39, 32)).compat_ret
        assert s
        assert r[0][0] is w
        assert claripy.backends.vsa.min(r[0][1]) == 0
        all_vals = r[0][1]._model_vsa.eval(1000)
        assert set(all_vals) == {4294967295, 0, 1}

    def test_overflow(self):
        x = claripy.BVS("x", 32)

        # x + 10 <= 20
        s, r = claripy.balancer.Balancer(claripy.backends.vsa, x + 10 <= claripy.BVV(20, 32)).compat_ret
        # mn,mx = claripy.backends.vsa.min(r[0][1]), claripy.backends.vsa.max(r[0][1])
        assert s
        assert r[0][0] is x
        assert set(claripy.backends.vsa.eval(r[0][1], 1000)) == {
            4294967286,
            4294967287,
            4294967288,
            4294967289,
            4294967290,
            4294967291,
            4294967292,
            4294967293,
            4294967294,
            4294967295,
            0,
            1,
            2,
            3,
            4,
            5,
            6,
            7,
            8,
            9,
            10,
        }

        # x - 10 <= 20
        s, r = claripy.balancer.Balancer(claripy.backends.vsa, x - 10 <= claripy.BVV(20, 32)).compat_ret
        assert s
        assert r[0][0] is x
        assert set(claripy.backends.vsa.eval(r[0][1], 1000)) == set(range(10, 31))

    def test_extract_zeroext(self):
        x = claripy.BVS("x", 8)
        expr = claripy.Extract(31, 0, claripy.ZeroExt(56, x)) <= claripy.BVV(0xE, 32)
        s, r = claripy.balancer.Balancer(claripy.backends.vsa, expr).compat_ret

        assert s is True
        assert len(r) == 1
        assert r[0][0] is x

    def test_complex_case_0(self):
        """
        <Bool ((0#48 .. Reverse(unconstrained_read_69_16)) << 0x30) <= 0x40000000000000>

        Created by VEX running on the following x86_64 assembly:
        cmp     word ptr [rdi], 40h
        ja      skip
        """

        x = claripy.BVS("x", 16)
        expr = (claripy.ZeroExt(48, claripy.Reverse(x)) << 0x30) <= 0x40000000000000

        s, r = claripy.balancer.Balancer(claripy.backends.vsa, expr).compat_ret

        assert s
        assert r[0][0] is x
        assert set(claripy.backends.vsa.eval(r[0][1], 1000)) == set(range(0, 65 * 0x100, 0x100))

    def test_complex_case_1(self):
        """
        <Bool (0#31 .. (if (0x8 < (0#32 .. (0xfffffffe + reg_284_0_32{UNINITIALIZED}))) then 1 else 0)) == 0x0>

        Created by VEX running on the following S390X assembly:
        0x40062c:       ahik    %r2, %r11, -2
        0x400632:       clijh   %r2, 8, 0x40065c

        IRSB {
        t0:Ity_I32 t1:Ity_I32 t2:Ity_I32 t3:Ity_I32 t4:Ity_I32 t5:Ity_I32
        t6:Ity_I64 t7:Ity_I64 t8:Ity_I64 t9:Ity_I64 t10:Ity_I32 t11:Ity_I1
        t12:Ity_I64 t13:Ity_I64 t14:Ity_I64 t15:Ity_I32 t16:Ity_I1

        00 | ------ IMark(0x40062c, 6, 0) ------
        01 | t0 = GET:I32(r11_32)
        02 | t1 = Add32(0xfffffffe,t0)
        03 | PUT(352) = 0x0000000000000003
        04 | PUT(360) = 0xfffffffffffffffe
        05 | t13 = 32Sto64(t0)
        06 | t7 = t13
        07 | PUT(368) = t7
        08 | PUT(376) = 0x0000000000000000
        09 | PUT(r2_32) = t1
        10 | PUT(ia) = 0x0000000000400632
        11 | ------ IMark(0x400632, 6, 0) ------
        12 | t14 = 32Uto64(t1)
        13 | t8 = t14
        14 | t16 = CmpLT64U(0x0000000000000008,t8)
        15 | t15 = 1Uto32(t16)
        16 | t10 = t15
        17 | t11 = CmpNE32(t10,0x00000000)
        18 | if (t11) { PUT(ia) = 0x40065c; Ijk_Boring }
        NEXT: PUT(ia) = 0x0000000000400638; Ijk_Boring
        }
        """

        x = claripy.BVS("x", 32)
        expr = claripy.ZeroExt(
            31, claripy.If(claripy.BVV(0x8, 32) < claripy.BVV(0xFFFFFFFE, 32) + x, claripy.BVV(1, 1), claripy.BVV(0, 1))
        ) == claripy.BVV(0, 32)
        s, r = claripy.balancer.Balancer(claripy.backends.vsa, expr).compat_ret

        assert s is True
        assert len(r) == 1
        assert r[0][0] is x

    def test_complex_case_2(self):
        x = claripy.BVS("x", 32)
        expr = claripy.ZeroExt(
            31, claripy.If(claripy.BVV(0xC, 32) < x, claripy.BVV(1, 1), claripy.BVV(0, 1))
        ) == claripy.BVV(0, 32)
        s, r = claripy.balancer.Balancer(claripy.backends.vsa, expr).compat_ret

        assert s is True
        assert len(r) == 1
        assert r[0][0] is x


if __name__ == "__main__":
    unittest.main()



# pylint: disable=missing-class-docstring,no-self-use
from __future__ import annotations

import unittest

import claripy
from claripy.bv import BVV, Concat, Extract, SDiv, SignExt, ZeroExt
from claripy.errors import ClaripyTypeError, ClaripyValueError, ClaripyZeroDivisionError


class TestBv(unittest.TestCase):
    def test_bv(self):
        a = BVV(1, 8)
        b = BVV(2, 8)
        assert a | b == 3
        assert a & b == 0
        assert a / b == 0
        assert b * b == 4
        assert a.signed == a.value
        assert a + 8 == 9

        c = BVV(128, 8)
        assert c.signed == -128

        d = BVV(255, 8)
        assert Extract(1, 0, d) == 3
        assert SignExt(8, d).value == 2**16 - 1
        assert ZeroExt(8, d).size() == 16
        assert ZeroExt(8, d).value == 255

        e = BVV(0b1010, 4)
        f = BVV(0b11, 2)
        assert Concat(e, e, e, e) == 0b1010101010101010
        assert Concat(e, f, f) == 0b10101111

        # test that __truediv__ is unsigned
        assert BVV(5, 8) / BVV(254, 8) == 0
        assert SDiv(BVV(5, 8), BVV(-2, 8)) == -2

        zero = BVV(0, 8)

        assert -a == 255
        assert ~a == 254
        assert -zero == 0
        assert ~zero == 255

    def test_get_byte(self):
        a = claripy.BVV(0xABCDEF12, 32)
        assert a.get_byte(0).args[0] == 0xAB
        assert a.get_byte(1).args[0] == 0xCD

        try:
            assert a.get_byte(4).args[0]
        except ValueError:
            pass
        else:
            assert False, "get_byte(4) did not raise an exception"

        b = claripy.BVV(0x1, 2)
        assert b.get_byte(0).args[0] == 0x1
        assert b.get_byte(0).args[1] == 8

    def test_get_bytes(self):
        a = claripy.BVV(0xABCDEF12, 32)
        assert a.get_bytes(0, 1).args[0] == 0xAB
        assert a.get_bytes(0, 2).args[0] == 0xABCD

        assert a.get_bytes(1, 1).args[0] == 0xCD

        try:
            assert a.get_bytes(4, 1).args[0]
        except ValueError:
            pass
        else:
            assert False, "get_bytes(4, 1) did not raise an exception"

        assert a.get_bytes(0, 0).args[0] == 0
        assert a.get_bytes(0, 0).args[1] == 0

        b = claripy.BVV(0x3F0, 10)
        assert b.get_bytes(0, 1).args[0] == 0x3
        assert b.get_bytes(0, 1).args[1] == 8
        assert b.get_bytes(0, 2).args[0] == 0x3F0
        assert b.get_bytes(0, 2).args[1] == 16
        assert b.get_bytes(1, 1).args[0] == 0xF0
        assert b.get_bytes(1, 1).args[1] == 8

    def test_zero_length(self):
        a = BVV(1, 8)
        b = BVV(0, 0)
        assert Concat(a, b) == 1

        self.assertRaises(ClaripyTypeError, lambda: a + b)
        self.assertRaises(ClaripyTypeError, lambda: a - b)
        self.assertRaises(ClaripyTypeError, lambda: a * b)
        self.assertRaises(ClaripyTypeError, lambda: a / b)

    def test_zero_division_errors(self):
        a = BVV(1, 32)
        b = BVV(0, 32)

        def _check_exception(va, vb, op):
            try:
                if hasattr(va, op):
                    getattr(va, op)(vb)
                elif hasattr(claripy, op):
                    getattr(claripy, op)(va, vb)
                else:
                    raise Exception(f"Operation {op} does not exist.")
                raise Exception("The expected exception is not raised.")
            except Exception as ex:  # pylint:disable=broad-except
                assert type(ex) is ClaripyZeroDivisionError

        _check_exception(a, b, "__truediv__")
        _check_exception(a, b, "__floordiv__")
        _check_exception(a, b, "__mod__")
        _check_exception(b, a, "__rtruediv__")
        _check_exception(b, a, "__rfloordiv__")
        _check_exception(b, a, "__rmod__")
        _check_exception(a, b, "SMod")
        _check_exception(a, b, "SDiv")

    def test_type_errors(self):
        self.assertRaises(TypeError, lambda: claripy.BVV(None))
        self.assertRaises(TypeError, lambda: claripy.BVV(3))
        self.assertRaises(TypeError, lambda: claripy.BVV(1.2))
        self.assertRaises(TypeError, lambda: claripy.BVV(b"asdf", b"qwer"))
        self.assertRaises(ClaripyValueError, lambda: claripy.BVV(b"asdf", 8))

        self.assertRaises(TypeError, lambda: claripy.BVS(None, None))
        self.assertRaises(TypeError, lambda: claripy.BVS(None, 3))
        self.assertRaises(TypeError, lambda: claripy.BVS(3, 3))
        self.assertRaises(TypeError, lambda: claripy.BVS(3, None))
        self.assertRaises(TypeError, lambda: claripy.BVS("asdf", None))


if __name__ == "__main__":
    unittest.main()



# pylint: disable=missing-class-docstring,no-self-use
from __future__ import annotations

import unittest

import claripy
from claripy.backends.backend import Backend


class AnnotationA(claripy.Annotation):
    def __init__(self, letter, number):
        self.letter = letter
        self.number = number
        claripy.Annotation.__init__(self)


class AnnotationB(AnnotationA):
    @property
    def eliminatable(self):
        return False

    @property
    def relocatable(self):
        return False


class AnnotationC(AnnotationA):
    def __init__(self, letter, number):
        super().__init__(letter, number)
        self._relocatable = True

    @property
    def eliminatable(self):
        return False

    @property
    def relocatable(self):
        return self._relocatable

    def relocate(self, src, dst):
        return AnnotationC(self.letter, self.number + 1)


class AnnotationD(claripy.Annotation):
    @property
    def eliminatable(self):
        return False

    @property
    def relocatable(self):
        return True


class BackendA(Backend):
    def __init__(self):
        super().__init__(self)
        self._op_expr["BVV"] = lambda e, result=None: e.args[0]

    def apply_annotation(self, o, a):
        return o + a.number


class TestAnnotation(unittest.TestCase):
    def test_backend(self):
        x = claripy.BVV(10, 32).annotate(AnnotationA("a", 1))
        assert BackendA().convert(x) == 11

    def test_simplification(self):
        x = claripy.BVS("x", 32).annotate(AnnotationA("a", 1))
        y = x ^ x
        assert y.depth == 1
        assert len(y.annotations) == 0

        x = claripy.BVS("x", 32).annotate(AnnotationB("a", 1))
        y = x ^ x
        assert y.depth == 2

        x = claripy.BVS("x", 32).annotate(AnnotationC("a", 1))
        y = x ^ x
        assert y.depth == 1
        assert len(y.annotations) == 1
        assert y.annotations[0].number == 2

    def test_missing_annotations_from_simplification(self):
        relocatable_anno = AnnotationC("a", 2)

        x0 = claripy.BVS("x", 32)
        x1 = claripy.BVV(24, 32)
        k = (x1 + x0).annotate(relocatable_anno)

        x3 = claripy.simplify(k)

        assert len(x3.annotations) == 1

    def test_annotations(self):
        x = claripy.BVS("x", 32) + 1
        xx = x._apply_to_annotations(lambda a: a)
        assert x is xx

        a1 = AnnotationA("a", 1)
        a2 = AnnotationA("a", 1)

        x1 = x.annotate(a1)
        x2 = x1.annotate(a2)
        x2a = x.annotate(a1, a2)
        x3 = x2.remove_annotation(a1)
        x4 = x3.remove_annotation(a2)
        x5 = x2.remove_annotations({a1, a2})

        assert x.variables == x1.variables
        assert x.variables == x2.variables
        assert x.variables == x2a.variables
        assert x.variables == x3.variables
        assert x.variables == x4.variables
        assert x.variables == x5.variables

        assert x is not x1
        assert x is not x2
        assert x is not x3
        assert x1 is not x2
        assert x1 is not x3
        assert x2 is not x3
        assert x2 is x2a
        assert x is x4
        assert x is x5

        assert x.op == x1.op
        assert x.annotations == ()
        assert x1.annotations == (a1,)
        assert x2.annotations == (a1, a2)
        assert x3.annotations == (a2,)

        assert claripy.backends.z3.convert(x).eq(claripy.backends.z3.convert(x3))

        const = claripy.BVV(1, 32)
        consta = const.annotate(AnnotationB("a", 0))
        const1 = consta + 1
        const1a = const1.annotate(AnnotationB("b", 1))
        const2 = const1a + 1
        # const2 should be (const1a + 1), instead of (1 + 1 + 1)
        # the flatten simplifier for __add__ should not be applied as
        # AnnotationB is not relocatable (and not eliminatable)
        assert const2.depth == 3

    def test_eagerness(self):
        x = claripy.BVV(10, 32).annotate(AnnotationD())
        y = x + 1
        assert y.annotations == x.annotations

    def test_ast_hash_should_consider_relocatable_annotations(self):
        relocatable_anno = AnnotationC("a", 2)
        const = claripy.BVV(1337, 32)
        x0 = claripy.BVS("x", 32).annotate(relocatable_anno)
        y0 = claripy.Concat(x0, const)

        # make the annotation not relocatable
        # this is of course a hack, but it can demonstrate the problem
        relocatable_anno._relocatable = False
        x0._relocatable_annotations = frozenset()

        y1 = claripy.Concat(x0, const)

        assert len(y0.annotations) == 1
        assert len(y1.annotations) == 0
        assert y0._hash != y1._hash

    def test_clear_annotations(self):
        x = claripy.BVS("x", 32).annotate(AnnotationA("a", 1))
        y = x.clear_annotations()
        assert len(y.annotations) == 0

        x2 = claripy.BVS("x", 32).annotate(AnnotationA("a", 1)).annotate(AnnotationA("b", 2))
        y2 = x2.clear_annotations()
        assert len(y2.annotations) == 0

    def test_remove_relocatable_annotations(self):
        relocatable_anno = AnnotationC("a", 2)
        const = claripy.BVV(1337, 32)

        x0 = claripy.BVS("x", 32).annotate(relocatable_anno)
        y0 = claripy.Concat(x0, const)
        assert len(y0.annotations) == 1
        assert y0.annotations == (relocatable_anno,)

        y1 = y0.remove_annotation(relocatable_anno)

        assert len(y1.annotations) == 0

    def test_duplicated_annotations_from_makelike(self):
        relocatable_anno = AnnotationC("a", 2)

        x0 = claripy.BVS("x", 32).annotate(relocatable_anno)
        x1 = claripy.BVV(24, 32)

        # make_like() should not re-apply child annotations if the child is the make_like target
        x2 = x0 + x1
        assert len(x2.annotations) == 1

        # simplify() should not re-apply annotations since annotations are kept during the simplification process by
        # make_like().
        x3 = claripy.simplify(x0 + x1)
        assert len(x3.annotations) == 1


if __name__ == "__main__":
    unittest.main()



# pylint: disable=missing-class-docstring,no-self-use
from __future__ import annotations

import unittest

from claripy.vsa import StridedInterval


def check_si_fields(si, stride, lb, ub):
    if si.stride != stride:
        return False
    if si.lower_bound != lb:
        return False
    return si.upper_bound == ub


class TestStridedInterval(unittest.TestCase):
    def test_smart_join(self):
        s1 = StridedInterval(bits=4, stride=3, lower_bound=9, upper_bound=12)
        s2 = StridedInterval(bits=4, stride=3, lower_bound=0, upper_bound=3)
        j = StridedInterval.pseudo_join(s1, s2)
        u = StridedInterval.least_upper_bound(s1, s2)
        assert check_si_fields(u, 3, 0, 12)
        assert check_si_fields(j, 3, 0, 12)

        s1 = StridedInterval(bits=4, stride=0, lower_bound=8, upper_bound=8)
        s2 = StridedInterval(bits=4, stride=1, lower_bound=14, upper_bound=15)
        s3 = StridedInterval(bits=4, stride=1, lower_bound=0, upper_bound=4)
        u = StridedInterval.least_upper_bound(s1, s2, s3)
        assert check_si_fields(u, 1, 14, 8)

        s1 = StridedInterval(bits=4, stride=3, lower_bound=2, upper_bound=8)
        s2 = StridedInterval(bits=4, stride=0, lower_bound=1, upper_bound=1)
        j = StridedInterval.pseudo_join(s1, s2)
        u = StridedInterval.least_upper_bound(s1, s2)
        assert check_si_fields(u, 3, 2, 1)
        assert check_si_fields(j, 3, 2, 1)


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations


class ConcreteHandlerMixin:
    def eval(self, e, n, extra_constraints=(), exact=None):
        c = self._concrete_value(e)
        if c is not None:
            return (c,)
        else:
            return super().eval(e, n, extra_constraints=extra_constraints, exact=exact)

    def batch_eval(self, exprs, n, extra_constraints=(), exact=None):
        concrete_exprs = [self._concrete_value(e) for e in exprs]
        symbolic_exprs = [e for e, c in zip(exprs, concrete_exprs) if c is None]

        if len(symbolic_exprs) == 0:
            return [concrete_exprs]

        symbolic_results = map(
            list, super().batch_eval(symbolic_exprs, n, extra_constraints=extra_constraints, exact=exact)
        )
        return [tuple((c if c is not None else r.pop(0)) for c in concrete_exprs) for r in symbolic_results]

    def max(self, e, extra_constraints=(), signed=False, exact=None):
        c = self._concrete_value(e)
        if c is not None:
            return c
        else:
            return super().max(e, extra_constraints=extra_constraints, signed=signed, exact=exact)

    def min(self, e, extra_constraints=(), signed=False, exact=None):
        c = self._concrete_value(e)
        if c is not None:
            return c
        else:
            return super().min(e, extra_constraints=extra_constraints, signed=signed, exact=exact)

    def solution(self, e, v, extra_constraints=(), exact=None):
        ce = self._concrete_value(e)
        cv = self._concrete_value(v)

        if ce is None or cv is None:
            return super().solution(e, v, extra_constraints=extra_constraints, exact=exact)
        else:
            return ce == cv

    def is_true(self, e, extra_constraints=(), exact=None):
        c = self._concrete_value(e)
        if c is not None:
            return c
        else:
            return super().is_true(e, extra_constraints=extra_constraints, exact=exact)

    def is_false(self, e, extra_constraints=(), exact=None):
        c = self._concrete_value(e)
        if c is not None:
            return not c
        else:
            return super().is_false(e, extra_constraints=extra_constraints, exact=exact)



# pylint: disable=missing-class-docstring,no-self-use
from __future__ import annotations

import unittest

import claripy


class TestRegressions(unittest.TestCase):
    def test_issue16(self):
        s = claripy.SolverComposite()

        c = claripy.BVS("test", 32)
        s.add(c[7:0] != 0)

        assert s.satisfiable()
        s.add(c == 0)

        assert not s.satisfiable(extra_constraints=[claripy.BVS("lol", 32) == 0])
        assert not s.satisfiable()


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import logging
import threading
from typing import TYPE_CHECKING, Any, TypeVar, overload

from claripy.ast.bv import SGE, SLE, UGE, ULE
from claripy.backend_manager import backends
from claripy.errors import BackendError, ClaripyFrontendError, UnsatError

from .constrained_frontend import ConstrainedFrontend

if TYPE_CHECKING:
    from collections.abc import Iterable

    from claripy.ast.bool import Bool
    from claripy.ast.bv import BV
    from claripy.ast.fp import FP

l = logging.getLogger("claripy.frontends.full_frontend")

T = TypeVar("T", bound="FullFrontend")


class FullFrontend(ConstrainedFrontend):
    def __init__(self, solver_backend, timeout=None, max_memory=None, track=False, **kwargs):
        ConstrainedFrontend.__init__(self, **kwargs)
        self._track = track
        self._solver_backend = solver_backend
        self.timeout = timeout if timeout is not None else 300000
        self.max_memory = max_memory
        self._tls = threading.local()
        self._to_add = []

    def _blank_copy(self, c):
        super()._blank_copy(c)
        c._track = self._track
        c._solver_backend = self._solver_backend
        c.timeout = self.timeout
        c.max_memory = self.max_memory
        c._tls = threading.local()
        c._to_add = []

    def _copy(self, c):
        super()._copy(c)
        c._track = self._track
        c._tls.solver = getattr(self._tls, "solver", None)  # pylint:disable=no-member
        c._to_add = list(self._to_add)

    #
    # Serialization support
    #

    def __getstate__(self):
        return (
            self._solver_backend.__class__.__name__,
            self.timeout,
            self.max_memory,
            self._track,
            super().__getstate__(),
        )

    def __setstate__(self, s):
        backend_name, self.timeout, self.max_memory, self._track, base_state = s
        self._solver_backend = backends._backends_by_type[backend_name]
        # self._tls = None
        self._tls = threading.local()
        self._to_add = []
        super().__setstate__(base_state)

    #
    # Frontend Creation
    #

    def _get_solver(self):
        if getattr(self._tls, "solver", None) is None:
            self._tls.solver = self._solver_backend.solver(timeout=self.timeout, max_memory=self.max_memory)
            self._add_constraints()
        elif self._finalized and len(self._to_add) > 0:
            if not hasattr(self._solver_backend, "clone_solver") or self._solver_backend.reuse_z3_solver:
                # this function may return a cached solver
                self._tls.solver = self._solver_backend.solver(timeout=self.timeout, max_memory=self.max_memory)
            else:
                self._tls.solver = self._solver_backend.clone_solver(self._tls.solver)
            self._add_constraints()

        if len(self._to_add) > 0:
            self._add_constraints()

        solver = self._tls.solver
        if self._solver_backend.reuse_z3_solver:
            # we must re-add all constraints
            self._add_constraints()
        return solver

    def _add_constraints(self):
        self._solver_backend.add(self._tls.solver, self.constraints, track=self._track)
        self._to_add = []

    #
    # Constraint management
    #

    def _add(self, constraints, invalidate_cache=True):
        to_add = ConstrainedFrontend._add(self, constraints)
        self._to_add += to_add
        return to_add

    def simplify(self):
        ConstrainedFrontend.simplify(self)

        # TODO: should we do this?
        self._tls.solver = None
        self._to_add = []

        return self.constraints

    def check_satisfiability(self, extra_constraints=(), exact=None) -> bool:
        try:
            return self._solver_backend.check_satisfiability(
                extra_constraints=extra_constraints, solver=self._get_solver(), model_callback=self._model_hook
            )
        except BackendError as e:
            raise ClaripyFrontendError("Backend error during solve") from e

    def satisfiable(self, extra_constraints: Iterable[Bool] = (), exact: bool | None = None) -> bool:
        try:
            return self._solver_backend.satisfiable(
                extra_constraints=extra_constraints, solver=self._get_solver(), model_callback=self._model_hook
            )
        except BackendError as e:
            raise ClaripyFrontendError("Backend error during solve") from e

    @overload
    def eval(
        self, e: BV, n: int, extra_constraints: tuple[Bool, ...] = ..., exact: Bool | None = ...
    ) -> tuple[int, ...]: ...

    @overload
    def eval(
        self, e: Bool, n: int, extra_constraints: tuple[Bool, ...] = ..., exact: Bool | None = ...
    ) -> tuple[bool, ...]: ...

    @overload
    def eval(
        self, e: FP, n: int, extra_constraints: tuple[Bool, ...] = ..., exact: Bool | None = ...
    ) -> tuple[float, ...]: ...

    def eval(self, e, n, extra_constraints=(), exact=None) -> tuple[Any, ...]:
        try:
            results = tuple(
                self._solver_backend.eval(
                    e,
                    n,
                    extra_constraints=extra_constraints,
                    solver=self._get_solver(),
                    model_callback=self._model_hook,
                )
            )
            if len(results) == 0:
                raise UnsatError("unsat")
            return results
        except BackendError as exc:
            raise ClaripyFrontendError("Backend error during eval") from exc

    # this is technically wrong but we cannot do better because python doesn't have associated types
    # if you need another overload, add one
    @overload
    def batch_eval(
        self, exprs: Iterable[BV], n: int, extra_constraints: tuple[Bool, ...] = ..., exact: bool | None = ...
    ) -> list[tuple[int, ...]]: ...

    @overload
    def batch_eval(
        self, exprs: Iterable[Bool], n: int, extra_constraints: tuple[Bool, ...] = ..., exact: bool | None = ...
    ) -> list[tuple[bool, ...]]: ...

    @overload
    def batch_eval(
        self, exprs: Iterable[FP], n: int, extra_constraints: tuple[Bool, ...] = ..., exact: bool | None = ...
    ) -> list[tuple[float, ...]]: ...

    def batch_eval(self, exprs, n, extra_constraints=(), exact=None):
        try:
            results = self._solver_backend.batch_eval(
                exprs,
                n,
                extra_constraints=extra_constraints,
                solver=self._get_solver(),
                model_callback=self._model_hook,
            )
            if len(results) == 0:
                raise UnsatError("unsat")
            return results
        except BackendError as e:
            raise ClaripyFrontendError("Backend error during batch_eval") from e

    @overload
    def max(
        self, e: BV, extra_constraints: tuple[Bool, ...] = ..., signed: bool = ..., exact: Bool | None = ...
    ) -> int: ...

    @overload
    def max(
        self, e: Bool, extra_constraints: tuple[Bool, ...] = ..., signed: bool = ..., exact: Bool | None = ...
    ) -> bool: ...

    @overload
    def max(
        self, e: FP, extra_constraints: tuple[Bool, ...] = ..., signed: bool = ..., exact: Bool | None = ...
    ) -> float: ...

    def max(self, e, extra_constraints=(), signed=False, exact=None):
        if not self.satisfiable(extra_constraints=extra_constraints):
            raise UnsatError("Unsat during _max()")

        l.debug("Frontend.max() with %d extra_constraints", len(extra_constraints))

        # pylint: disable=unsubscriptable-object
        two = self.eval(e, 2, extra_constraints=extra_constraints)
        if len(two) == 0:
            raise UnsatError("unsat during max()")
        if len(two) == 1:
            return two[0]

        if signed:
            c = (*tuple(extra_constraints), SGE(e, two[0]), SGE(e, two[1]))
        else:
            c = (*tuple(extra_constraints), UGE(e, two[0]), UGE(e, two[1]))
        try:
            return self._solver_backend.max(
                e,
                extra_constraints=c,
                solver=self._get_solver(),
                model_callback=self._model_hook,
                signed=signed,
            )
        except BackendError as exc:
            raise ClaripyFrontendError("Backend error during max") from exc

    @overload
    def min(
        self, e: BV, extra_constraints: tuple[Bool, ...] = ..., signed: bool = ..., exact: Bool | None = ...
    ) -> int: ...

    @overload
    def min(
        self, e: Bool, extra_constraints: tuple[Bool, ...] = ..., signed: bool = ..., exact: Bool | None = ...
    ) -> bool: ...

    @overload
    def min(
        self, e: FP, extra_constraints: tuple[Bool, ...] = ..., signed: bool = ..., exact: Bool | None = ...
    ) -> float: ...

    def min(self, e, extra_constraints=(), signed=False, exact=None):
        if not self.satisfiable(extra_constraints=extra_constraints):
            raise UnsatError("Unsat during _min()")

        l.debug("Frontend.min() with %d extra_constraints", len(extra_constraints))

        # pylint: disable=unsubscriptable-object
        two = self.eval(e, 2, extra_constraints=extra_constraints)
        if len(two) == 0:
            raise UnsatError("unsat during min()")
        if len(two) == 1:
            return two[0]

        if signed:
            c = (*tuple(extra_constraints), SLE(e, two[0]), SLE(e, two[1]))
        else:
            c = (*tuple(extra_constraints), ULE(e, two[0]), ULE(e, two[1]))
        try:
            return self._solver_backend.min(
                e,
                extra_constraints=c,
                solver=self._get_solver(),
                model_callback=self._model_hook,
                signed=signed,
            )
        except BackendError as exc:
            raise ClaripyFrontendError("Backend error during min") from exc

    @overload
    def solution(self, e: BV, v: int, extra_constraints: tuple[Bool, ...] = ..., exact: Bool | None = ...) -> bool: ...

    @overload
    def solution(
        self, e: Bool, v: bool, extra_constraints: tuple[Bool, ...] = ..., exact: Bool | None = ...
    ) -> bool: ...

    @overload
    def solution(
        self, e: FP, v: float, extra_constraints: tuple[Bool, ...] = ..., exact: Bool | None = ...
    ) -> bool: ...

    def solution(self, e, v, extra_constraints=(), exact=None):
        try:
            return self._solver_backend.solution(
                e, v, extra_constraints=extra_constraints, solver=self._get_solver(), model_callback=self._model_hook
            )
        except BackendError as exc:
            raise ClaripyFrontendError("Backend error during solution") from exc

    def is_true(self, e: Bool, extra_constraints: tuple[Bool, ...] = (), exact: bool | None = None) -> bool:
        return e.is_true()

    def is_false(self, e: Bool, extra_constraints: tuple[Bool, ...] = (), exact: bool | None = None) -> bool:
        return e.is_false()

    def unsat_core(self, extra_constraints: tuple[Bool, ...] = ()) -> Iterable[Bool]:
        if self.satisfiable(extra_constraints=extra_constraints):
            # all constraints are satisfied
            return ()

        unsat_core = self._solver_backend.unsat_core(self._get_solver())

        return tuple(unsat_core)

    #
    # Serialization and such.
    #

    def downsize(self) -> None:
        ConstrainedFrontend.downsize(self)
        self._tls.solver = None
        self._to_add = []

    #
    # Merging and splitting
    #

    def merge(self: T, others, merge_conditions, common_ancestor=None) -> tuple[bool, T]:
        return (
            self._solver_backend.__class__.__name__ == "BackendZ3",
            ConstrainedFrontend.merge(self, others, merge_conditions, common_ancestor=common_ancestor)[1],
        )

    #
    # Default model hook
    #

    def _model_hook(self, m):  # pylint:disable=unused-argument,no-self-use
        return None



from __future__ import annotations


class SimplifyHelperMixin:
    def max(self, e, extra_constraints=(), signed=False, exact=None):
        self.simplify()
        return super().max(e, extra_constraints=extra_constraints, signed=signed, exact=exact)

    def min(self, e, extra_constraints=(), signed=False, exact=None):
        self.simplify()
        return super().min(e, extra_constraints=extra_constraints, signed=signed, exact=exact)

    def eval(self, e, n, extra_constraints=(), exact=None):
        if n > 1:
            self.simplify()
        return super().eval(e, n, extra_constraints=extra_constraints, exact=exact)

    def batch_eval(self, exprs, n, extra_constraints=(), exact=None):
        if n > 1:
            self.simplify()
        return super().batch_eval(exprs, n, extra_constraints=extra_constraints, exact=exact)
